import { LRLanguage, LanguageSupport, LanguageDescription } from "@codemirror/language";
import { MetaStore, Hover, Problem, Position, RainDocument, CompletionItem, TextDocumentItem, RainLanguageServices } from "@rainlanguage/dotrain";
import { Facet, Text, TransactionSpec } from "@codemirror/state";
import { ViewPlugin, ViewUpdate, PluginValue, EditorView, Tooltip } from "@codemirror/view";
import { Diagnostic } from "@codemirror/lint";
import { TextDocument } from "vscode-languageserver-textdocument";
import { CompletionContext, CompletionResult, Completion } from "@codemirror/autocomplete";
/**
 * @public Rainlang syntax highlighter
 */
declare const RainlangLR: LRLanguage;
/**
 * @public A function for calling the native parser and get back its errors as Problems
 */
type NativeParserCallback = (dotrain: RainDocument) => Promise<Problem[]>;
/**
 * @public Converts TextDocument to TextDocumentItem
 */
declare function toTextDocumentItem(textDocument: TextDocument): TextDocumentItem;
// const useLast = (values: readonly any[]) => values.reduce((_, v) => v, '');
// const documentUri = Facet.define<string, string>({ combine: useLast });
// const languageId = Facet.define<string, string>({ combine: useLast });
/**
 * @public A plugin that manages languages services processes.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 */
declare class RainLanguageServicesPlugin implements PluginValue {
    view: EditorView;
    callback?: NativeParserCallback | undefined;
    readonly uri = "file:///untitled.rain";
    readonly languageId = "rainlang";
    metaStore: MetaStore;
    version: number;
    langServices: RainLanguageServices;
    private textDocument;
    /**
     * @public constructor of the class
     * @param view - The editor view instance
     * @param metaStore - (optional) The meta store object to use
     * @param callback - (optional) The callback to run fork calls on native parser,
     * this will be called everytime there are no active diagnostics, so as a result,
     * errors from calling onchain parser will be shown as diagnostics
     */
    constructor(view: EditorView, metaStore?: MetaStore, callback?: NativeParserCallback | undefined);
    // update the instance of TexTdocument and RainDocument
    update(update: ViewUpdate): void;
    destroy(): void;
    /**
     * @public Handles calls for hover tooltip
     */
    handleHoverTooltip(offset: number): Promise<Tooltip | null>;
    /**
     * @public Handles calls for code completion
     */
    handleCompletion(context: CompletionContext): Promise<CompletionResult | null>;
    /**
     * @public Handles calls for docuement diagnostics
     */
    processDiagnostics(text: string, version: number): Promise<[
        TransactionSpec,
        number
    ]>;
}
/**
 * @public Defines the raw language services callbacks signature
 */
type RawLanguageServicesCallbacks = {
    diagnostics: (textDocument: TextDocumentItem) => Promise<Problem[]>;
    hover?: (textDocument: TextDocumentItem, position: Position) => Promise<Hover | null>;
    completion?: (textDocument: TextDocumentItem, position: Position) => Promise<CompletionItem[] | null>;
};
/**
 * @public A plugin that gets all lang services through provided callbacks.
 * So all the logic behind suppliying the lang services results can be out-sources
 * on the callbacks implementations.
 * Each callback is triggered when the services is on demand in the text editor
 * through use input/actions
 */
declare class RawRainLanguageServicesPlugin implements PluginValue {
    view: EditorView;
    callbacks: RawLanguageServicesCallbacks;
    readonly uri = "file:///untitled.rain";
    readonly languageId = "rainlang";
    version: number;
    private textDocument;
    constructor(view: EditorView, callbacks: RawLanguageServicesCallbacks);
    // update the instance of TexTdocument and RainDocument
    update(update: ViewUpdate): void;
    destroy(): void;
    /**
     * @public Handles calls for hover tooltip
     */
    handleHoverTooltip(offset: number): Promise<Tooltip | null>;
    /**
     * @public Handles calls for code completion
     */
    handleCompletion(context: CompletionContext): Promise<CompletionResult | null>;
    /**
     * @public Handles calls for docuement diagnostics
     */
    processDiagnostics(text: string, version: number): Promise<[
        TransactionSpec,
        number
    ]>;
}
/**
 * @public Converts LSP hover to codemirror hover tooltips.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param hover - The LSP hover item
 * @param pos - Offset of the textDocument to get the completion items for
 * @param end - The end offset of the requested hiver item
 * @returns Codemirror tooltip or null
 */
declare function getHoverTooltip(hover: Hover, pos: number, end?: number): Tooltip | null;
/**
 * @public Converts LSP CompletionItem to codemirror code completions.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param context - Completion context
 * @param completionItems - The LSP completion items
 * @returns Codemirror completion result or null
 */
declare function getCompletion(context: CompletionContext, completionItems: CompletionItem[]): CompletionResult | null;
/**
 * @public Converts RainDocument Problems to codemirror diagnostics
 * @param problems - RainDocument problems
 */
declare function getDiagnostics(problems: Problem[]): Diagnostic[];
/**
 * @public Converts a position to offset of a character in a codemirror text.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror text
 * @param pos - The position to convert
 * @returns - The offset of the provided position
 */
declare function posToOffset(doc: Text, pos: {
    line: number;
    character: number;
}): number;
/**
 * @public Converts offset of a character in a codemirror text to a position.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror text
 * @param offset - The offset to convert
 * @returns - The position of the provided character offset
 */
declare function offsetToPos(doc: Text, offset: number): {
    line: number;
    character: number;
};
/**
 * @public Constructs regexp out of completion items.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param completions - Codemirror completion items
 * @returns Regexp of completion items
 */
declare function prefixMatch(completions: Completion[]): RegExp[];
/**
 * @public Reduces an array to its last element
 * @param values - Values to reduce
 */
declare const useLast: (values: readonly any[]) => any;
/**
 * @public The facet used to store and retrive RainLanguageServicesPlugin instance from a EditorView.plugin().
 * By getting the facet value from the state associated with that EditoreView instance, you then can intract
 * with the plugin directly to get the Meta.Store instance and update its subgraph and metas manually, however,
 * this is not recommended as the meta storage and management is handled automatically by every meta hash that
 * is specified in an Rain document, so using this will not be needed in most of the cases unless some custom
 * behavior is intended.
 * This combination is provided for `rainlang` LanguageSupport class object and not for `RainlangExtention` class.
 *
 * @example
 * ```typescript
 * // get value of the facet
 * const rainViewPlugin = editorView.state.facet(RainLanguageServicesFacet);
 *
 * // get the plugin instance
 * const rainPlugin = editorView.plugin(rainViewPlugin);
 *
 * // update op meta from plugin
 * rainPlugin.getMetaStore();
 * ```
 */
declare const RainLanguageServicesFacet: Facet<ViewPlugin<RainLanguageServicesPlugin>, ViewPlugin<RainLanguageServicesPlugin>>;
/**
 * @public Configuraions options for rain language services
 */
type LanguageServicesConfig = {
    /**
     * if hover tooltips should be actiavted
     */
    hover?: boolean;
    /**
     * if code completion suggestions should be actiavted
     */
    completion?: boolean;
    /**
     * callback for getting native parser errors
     */
    callback?: NativeParserCallback;
    /**
     * The MetaStore to instantiate rain language services with
     */
    metaStore?: MetaStore;
};
/**
 * @public Rainlang as a codemirror extension
 * @example
 * ```typescript
 * // instantiate the extension that can be directly used as codemirror extension
 * const rainlangExtension = new RainlangExtention(options);
 *
 * // instantiate the extension with initial meta data
 * const rainlangExtension = new RainlangExtension();
 *
 * // to get the MetaStore instance of this extension instance
 * const metaStore = rainlangExtension.plugin.metaStore;
 * ```
 */
declare class RainlangExtension extends LanguageSupport {
    plugin: RainLanguageServicesPlugin | undefined;
    private hover;
    private completion;
    /**
     * @public Constructor of the RainlangCodemirror class
     * @param config - Options for language services to include
     */
    /**
     * @public Constructor of the RainlangCodemirror class
     * @param config - Options for language services to include
     */
    constructor(config?: LanguageServicesConfig);
}
/**
 * @public Raw Rainlang codemirror extension, that does everything through supplied callbacks
 */
declare class RawRainlangExtension extends LanguageSupport {
    plugin: RawRainLanguageServicesPlugin | undefined;
    private hover;
    private completion;
    /**
     * @public Constructor of the RainlangCodemirror class
     * @param callbacks - Raw language services callbacks
     */
    /**
     * @public Constructor of the RainlangCodemirror class
     * @param callbacks - Raw language services callbacks
     */
    constructor(callbacks: RawLanguageServicesCallbacks);
}
/**
 * @public
 * Provides Rainlang implementation for codemirror as LanguageSupport
 * This is the standard implementation of rainlang for codemirror.
 *
 * @example
 * ```typescript
 * const rainLanguageSupport = RainLanguageSupport(config);
 * ```
 * in order to get the Meta.Store instance, it should be done through getting the plugin instance
 * from the EditorView with providing it the result of getting the `RainLanguageServicesFacet`
 * facet and then calling `getMetaStore()` for it, however this is not recommended as meta management
 * is done automatically from specified meta hashes in a Rain document:
 * ```typescript
 * // get the `RainLanguageServicesFacet` value from editor view
 * const rainViewPlugin = view.state.facet(RainLanguageServicesFacet);
 *
 * // retrive the `RainLanguageServicesPlugin` instance
 * const rainPlugin = view.plugin(rainViewPlugin);
 *
 * // get the Meta.Store instance
 * rainPlugin.metaStore();
 * ```
 *
 * @param config - Options for language services to include
 * @param metaStore - (optional) The meta store object to use
 * @returns rainlang LanguageSupport
 */
declare function RainLanguageSupport(config?: LanguageServicesConfig): LanguageSupport;
/**
 * @public Provides LanguageDescription class object for rainlang.
 * Use `load` to instantiate rain LanguageSupport with initial meta data
 */
declare const RainlangDescription: LanguageDescription;
export { RainlangLR, MetaStore, NativeParserCallback, toTextDocumentItem, RainLanguageServicesPlugin, RawLanguageServicesCallbacks, RawRainLanguageServicesPlugin, getHoverTooltip, getCompletion, getDiagnostics, posToOffset, offsetToPos, prefixMatch, useLast, RainLanguageServicesFacet, LanguageServicesConfig, RainlangExtension, RawRainlangExtension, RainLanguageSupport, RainlangDescription };
export * from "@rainlanguage/dotrain";
export { TextDocument, TextEdit, TextDocumentContentChangeEvent } from "vscode-languageserver-textdocument";
